import { concat, asyncConcat } from "async-iterable-split";
import { TYPES } from "./constants.js";
class CommonIterator {
    constructor(type, source) {
        this.done = false;
        this.prefix = new Uint8Array([type]);
        this.source = Symbol.asyncIterator in source ? source[Symbol.asyncIterator]() : source[Symbol.iterator]();
    }
    async next() {
        if (this.done) {
            return { done: true, value: undefined };
        }
        let item;
        if (this.prefix !== undefined) {
            item = { done: false, value: this.prefix };
            this.prefix = undefined;
            return item;
        }
        item = await this.source.next();
        if (item.done) {
            this.done = true;
            return { done: true, value: undefined };
        }
        if (!(item.value instanceof Uint8Array)) {
            throw new Error("Invalid data");
        }
        return item;
    }
    async return() {
        this.done = true;
        if (this.source.return) {
            return this.source.return();
        }
        return { done: true, value: undefined };
    }
    async throw(error) {
        this.done = true;
        if (this.source.throw) {
            return this.source.throw(error);
        }
        return Promise.reject(error);
    }
}
class StreamIterator {
    constructor(type, source) {
        this.done = false;
        this.chunkData = [];
        this.chunkSize = 0;
        this.prefix = new Uint8Array([type]);
        this.source = Symbol.asyncIterator in source ? source[Symbol.asyncIterator]() : source[Symbol.iterator]();
    }
    async next() {
        var _a;
        if (this.done) {
            return { done: true, value: undefined };
        }
        let item;
        if (this.prefix !== undefined) {
            item = { done: false, value: this.prefix };
            this.prefix = undefined;
            return item;
        }
        if (this.chunkSize >= 65536) {
            return { done: false, value: this.flushChunk(false) };
        }
        if ((_a = this.remain) === null || _a === void 0 ? void 0 : _a.length) {
            item = { done: false, value: this.remain };
            this.remain = undefined;
        }
        else {
            item = await this.source.next();
        }
        if (item.done) {
            this.done = true;
            return { done: false, value: this.flushChunk(true) };
        }
        if (!(item.value instanceof Uint8Array)) {
            throw new Error("Invalid data");
        }
        const space = 4294967295 - this.chunkSize;
        if (space >= item.value.length) {
            this.chunkData.push(item.value);
            this.chunkSize += item.value.length;
        }
        else {
            this.chunkData.push(item.value.subarray(0, space));
            this.chunkSize = 4294967295;
            this.remain = item.value.subarray(space);
        }
        return this.next();
    }
    async return() {
        this.done = true;
        if (this.source.return) {
            return this.source.return();
        }
        return { done: true, value: undefined };
    }
    async throw(error) {
        this.done = true;
        if (this.source.throw) {
            return this.source.throw(error);
        }
        return Promise.reject(error);
    }
    flushChunk(done) {
        const dv = new DataView(new ArrayBuffer(4));
        dv.setUint32(0, this.chunkSize, false);
        this.chunkData.unshift(new Uint8Array(dv.buffer));
        if (this.chunkSize !== 0 && done) {
            this.chunkData.push(new Uint8Array([0, 0, 0, 0]));
        }
        const chunk = concat(this.chunkData);
        this.chunkData = [];
        this.chunkSize = 0;
        return chunk;
    }
}
function hasCircular(value, refs = new WeakSet()) {
    if (refs.has(value)) {
        return true;
    }
    if (value !== null && typeof value === "object") {
        refs.add(value);
        for (const key in value) {
            if (hasCircular(value[key], refs)) {
                return true;
            }
        }
        refs.delete(value);
    }
    return false;
}
export function generator(value) {
    if (value === null) {
        return {
            [Symbol.asyncIterator]() {
                return new CommonIterator(TYPES.NULL, []);
            },
        };
    }
    if (value === undefined) {
        return {
            [Symbol.asyncIterator]() {
                return new CommonIterator(TYPES.UNDEFINED, []);
            },
        };
    }
    if (typeof value === "boolean" || value instanceof Boolean) {
        return {
            [Symbol.asyncIterator]() {
                return new CommonIterator(TYPES.BOOLEAN, [new Uint8Array([+value])]);
            },
        };
    }
    if (typeof value === "number" || value instanceof Number) {
        return {
            [Symbol.asyncIterator]() {
                const dv = new DataView(new ArrayBuffer(8));
                dv.setFloat64(0, +value, false);
                return new CommonIterator(TYPES.NUMBER, [new Uint8Array(dv.buffer)]);
            },
        };
    }
    if (typeof value === "string" || value instanceof String) {
        return {
            [Symbol.asyncIterator]() {
                return new StreamIterator(TYPES.STRING, [new TextEncoder().encode("" + value)]);
            },
        };
    }
    if (value instanceof Uint8Array || Symbol.asyncIterator in value) {
        return {
            [Symbol.asyncIterator]() {
                return new StreamIterator(TYPES.STREAM, value instanceof Uint8Array ? [value] : value);
            },
        };
    }
    if (Object.prototype.toString.call(value) === "[object Array]") {
        return {
            [Symbol.asyncIterator]() {
                if (hasCircular(value)) {
                    throw new Error("Circular reference!");
                }
                return new StreamIterator(TYPES.ARRAY, multiGenerator(value));
            },
        };
    }
    if (Object.prototype.toString.call(value) === "[object Object]") {
        return {
            [Symbol.asyncIterator]() {
                if (hasCircular(value)) {
                    throw new Error("Circular reference!");
                }
                const flatted = [];
                for (const key in value) {
                    flatted.push(key);
                    flatted.push(value[key]);
                }
                return new StreamIterator(TYPES.OBJECT, multiGenerator(flatted));
            },
        };
    }
    throw new Error("Invalid data");
}
export function multiGenerator(values) {
    return {
        [Symbol.asyncIterator]() {
            let index = 0;
            let iterator;
            return {
                async next() {
                    if (index === values.length) {
                        return { done: true, value: undefined };
                    }
                    if (iterator === undefined) {
                        iterator = generator(values[index])[Symbol.asyncIterator]();
                    }
                    const item = await iterator.next();
                    if (!item.done) {
                        return item;
                    }
                    index++;
                    iterator = undefined;
                    return this.next();
                },
                async return() {
                    index = values.length;
                    if (iterator === null || iterator === void 0 ? void 0 : iterator.return) {
                        return iterator === null || iterator === void 0 ? void 0 : iterator.return();
                    }
                    return { done: true, value: undefined };
                },
                async throw(error) {
                    index = values.length;
                    if (iterator === null || iterator === void 0 ? void 0 : iterator.throw) {
                        return iterator === null || iterator === void 0 ? void 0 : iterator.throw(error);
                    }
                    return Promise.reject(error);
                },
            };
        },
    };
}
export function easyGenerator(value, maxSize) {
    return asyncConcat(generator(value), maxSize);
}
export function easyMultiGenerator(values, maxSize) {
    return asyncConcat(multiGenerator(values), maxSize);
}
