import { Splitable, createSplitable } from "async-iterable-split";
import { TYPES } from "./constants.js";
export async function parser(splitable, options = {}, paths = []) {
    const _options = {
        streamMaxSize: Infinity,
        stringMaxSize: Infinity,
        arrayMaxLength: Infinity,
        objectMaxLength: Infinity,
        maxDepth: Infinity,
        ...options,
    };
    if (paths.length > _options.maxDepth) {
        throw new Error(`Depth exceeded ${_options.maxDepth}`);
    }
    if (!(splitable instanceof Splitable)) {
        splitable = createSplitable(splitable);
    }
    const type = (await splitable.readEnoughSize(1))[0];
    if (type === TYPES.NULL) {
        return null;
    }
    if (type === TYPES.UNDEFINED) {
        return undefined;
    }
    if (type === TYPES.BOOLEAN) {
        const data = await splitable.readEnoughSize(1);
        if (data[0] === 0) {
            return false;
        }
        if (data[0] === 1) {
            return true;
        }
        throw new Error("Invalid data");
    }
    if (type === TYPES.NUMBER) {
        const data = await splitable.readEnoughSize(8);
        const dv = new DataView(data.buffer, data.byteOffset, data.byteLength);
        return dv.getFloat64(0, false);
    }
    if (type === TYPES.STREAM) {
        return splitable.readChunkedStream(_options.streamMaxSize);
    }
    if (type === TYPES.STRING) {
        let result = "";
        let size = 0;
        const subIterable = splitable.splitChunkedStream();
        const decoder = new TextDecoder("utf-8");
        for await (const chunk of subIterable) {
            if ((size += chunk.length) > _options.stringMaxSize) {
                throw new Error(`String size exceeded ${_options.stringMaxSize} bytes`);
            }
            result += decoder.decode(chunk, { stream: true });
        }
        return result;
    }
    if (type === TYPES.ARRAY) {
        let result = [];
        const subIterable = splitable.splitChunkedStream();
        const subSplitable = new Splitable(subIterable);
        while (await subSplitable.hasValue) {
            if (result.length >= _options.arrayMaxLength) {
                throw new Error(`Array length exceeded ${_options.arrayMaxLength}`);
            }
            const element = await parser(subSplitable, _options, [...paths, result.length]);
            result.push(element);
        }
        return result;
    }
    if (type === TYPES.OBJECT) {
        let result = {};
        let length = 0;
        const subIterable = splitable.splitChunkedStream();
        const subSplitable = new Splitable(subIterable);
        while (await subSplitable.hasValue) {
            if (length >= (_options === null || _options === void 0 ? void 0 : _options.objectMaxLength)) {
                throw new Error(`Object key value pairs count exceeded ${_options === null || _options === void 0 ? void 0 : _options.objectMaxLength}`);
            }
            if ((await subSplitable.readEnoughSize(1))[0] !== TYPES.STRING) {
                throw new Error("Invalid key");
            }
            const key = new TextDecoder("utf-8").decode(await subSplitable.readChunkedStream(_options.stringMaxSize));
            const value = await parser(subSplitable, _options, [...paths, key]);
            result[key] = value;
        }
        return result;
    }
    throw new Error("Invalid type");
}
